#!/usr/bin/env python
# encoding: utf-8
"""
desire2download.py

Copyright 2012 Stephen Holiday

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import BeautifulSoup
import mechanize
import os
import re
import urllib
import urllib2

import sys
reload(sys)
sys.setdefaultencoding("utf-8")


class AuthError(Exception):
    """Raised when login credentials fail."""
    pass


class Desire2Download(object):
    base_url = 'https://learn.uwaterloo.ca/d2l/home/6606'
    cas_login = 'https://cas.uwaterloo.ca/cas/login?service=http%3a%2f%2flearn.uwaterloo.ca%2fd2l%2forgtools%2fCAS%2fDefault.aspx'
    ping_url = 'http://jobminestats.appspot.com/Ping/ag5zfmpvYm1pbmVzdGF0c3IMCxIFUGl4ZWwYuRcM.gif'

    def __init__(self, username, password, ignore_re=None, retries=3, skip_existing=True):
        self.username = username
        self.password = password
        self.ignore_re = ignore_re
        self.retries = retries
        self.skip_existing = skip_existing

        self.br = mechanize.Browser(factory=mechanize.RobustFactory())
        self.br.addheaders = [('User-agent', 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_8_3)\
            AppleWebKit/537.31 (KHTML, like Gecko) Chrome/26.0.1410.65 Safari/537.31')]
        self.br.set_handle_refresh(mechanize._http.HTTPRefreshProcessor(), max_time=1)

        self.br.open(self.ping_url).read()

    def retry(f):
        """Decorator to retry upon timeout. D2L is slow."""
        def retry_it(self, *args, **kargs):
            attempts = 0
            while attempts < self.retries:
                try:
                    return f(self, *args, **kargs)
                except urllib2.URLError as e:
                    if isinstance(e.reason, socket.timeout):
                        attempts += 1
                        if attempts >= self.retries:
                            print "Timeout, out of retries."
                            raise(e)
                        print "Timeout, retrying..."
                    else: 
                        # Not a timeout, raise exception
                        print "Unknown exception:",e
                        raise(e)
        return retry_it

    @retry
    def login(self):
        print 'Logging In...'
        self.br.open(self.cas_login)
        self.br.select_form(nr=0)
        self.br['username'] = self.username
        self.br['password'] = self.password
        response = self.br.submit().read()
        if "Your userid and/or your password are incorrect" in response:
            raise AuthError("Your userid and/or your password are incorrect.")
        if "d2l" in response:
            print 'Logged In'
            return
        raise AuthError("Some other login error occured.")

    def get_course_links(self):
        print 'Finding courses...'
        links = []
        urls = []
        for link in self.br.links():
            if link.text is not None:
                matches = re.match('[A-Z]+ [0-9A-Za-z/\s]{2,45} - [A-Z][a-z]+ 20[0-9]{2}', link.text)
                if matches is not None and link.url not in urls:
                    links.append(link)
                    urls.append(link.url)
        return links

    def convert_bytes(self, bytes):
        """
            Stolen from http://www.5dollarwhitebox.org/drupal/node/84
        """
        bytes = float(bytes)
        if bytes >= 1099511627776:
            terabytes = bytes / 1099511627776
            size = '%.2fT' % terabytes
        elif bytes >= 1073741824:
            gigabytes = bytes / 1073741824
            size = '%.2fG' % gigabytes
        elif bytes >= 1048576:
            megabytes = bytes / 1048576
            size = '%.2fM' % megabytes
        elif bytes >= 1024:
            kilobytes = bytes / 1024
            size = '%.2fK' % kilobytes
        else:
            size = '%.2fb' % bytes
        return size

    @retry
    def get_course_documents(self, link, course_name):
        """Produce a tree of documents for the course.

        Args:
            link (str): A url to the course's page on d2l.
            course_name (str): The name of the course.

        Returns:
            A dict representing a tree:
            {
                'type': Either 'file' or 'dir',
                'name': A string.
                'url': Url to the file preview (if file).
                'children': A list of children nodes (if a dir).
            }
        """
        self.br.follow_link(link)                    # Go to course page
        link = self.br.links(text='Content').next()  # Get content link
        page = self.br.follow_link(link).read()      # Go to content page
        soup = BeautifulSoup.BeautifulSoup(page)

        ## Initial document tree
        document_tree = {
            'type': 'dir',
            'name': course_name,
            'children': []
        }

        for link in self.br.links(url_regex='/d2l/le/content/.+/viewContent/.+/View'):
            ou = re.search('/content/([0-9]+)/', link.url).group(1)
            tId = re.search('/viewContent/([0-9]+)/', link.url).group(1)
            link_href = 'https://learn.uwaterloo.ca/d2l/le/content/%s/%s/downloadfiles/DownloadTopic' % (ou, tId)
            
            document_tree['children'].append({
                'type': 'file',
                'name': link.text,
                'url': link_href
                })

        return document_tree

    def download_tree(self, root, _path=[]):
        """Downloads the entire file tree the

        Args:
            root: A dictionary containing the file tree.
            _path: A list representing the path (relative to current dir) to
                download to. Items in list are strings.
        """
        if root['type'] == 'dir':
            path = _path[:]
            path.append(root['name'])
            for node in root['children']:
                self.download_tree(node, path)
        else:
            path = '/'.join(map(lambda x: x.replace('/', '-'), _path))
            self.download_file(root['name'], root['url'], path)

    def download_file(self, title, url, path):
        """Downloads a file to the specified directory.

        Args:
            title (str): Name of the file.
            url (str): Address to the file preview page.
            path (str): Relative path of file to make.
        """
        try:
            os.makedirs(path)
        except OSError as e:
            if e.errno != 17:
                raise e
            pass

        try:
            ret = self.br.open(url)
        except Exception, e:
            print " X Error for %s [%s]" % (url, e)
            return
        
        content_disposition = next(x for x in ret.info().headers if 'Content-Disposition' in x)
        filename = re.findall("filename=(\S+)", content_disposition)[0]
        filename = urllib.unquote(filename).replace('"', '')

        path_and_filename = '%s/%s' % (path, filename.strip('/'))

        if os.path.isfile(path_and_filename) and self.skip_existing:  # TODO Can we make this smarter?
            print ' - %s (Already Saved)' % path_and_filename
            return

        try:
            print ' + %s' % path_and_filename
            self.br.retrieve(url, path_and_filename, self._progressBar)
        except KeyboardInterrupt:
            # delete the file on a keyboard interrupt
            if os.path.exists(path_and_filename):
                os.remove(path_and_filename)
            raise
        except urllib2.HTTPError, e:
            if e.code == 404:
                print " X File does not exist: %s" % file_name.strip('/')
            else:
                print " X HTTP error %s for: %s" % (e.code, file_name.strip('/'))
        except Exception, e:
            # otherwise raise the error
            if os.path.exists(path_and_filename):
                os.remove(path_and_filename)
            else:
                print url
                print e

    def _progressBar(self, blocknum, bs, size):
        """
            Stolen from https://github.com/KartikTalwar/Coursera/blob/master/coursera.py
        """
        if size > 0:
            if size % bs != 0:
                blockCount = size/bs + 1
            else:
                blockCount = size/bs

            fraction = blocknum*1.0/blockCount
            width    = 50

            stars    = '*' * int(width * fraction)
            spaces   = ' ' * (width - len(stars))
            progress = ' ' * 3 + '%s [%s%s] (%s%%)' % (self.convert_bytes(size), stars, spaces, int(fraction * 100))

            if fraction*100 < 100:
                sys.stdout.write(progress)

                if blocknum < blockCount:
                    sys.stdout.write('\r')
                else:
                    sys.stdout.write('\n')
            else:
                sys.stdout.write(' ' * int(width * 1.5) + '\r')
                sys.stdout.flush()




#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name='Desire2Download',
    version='0.2.0beta',
    description='Download all of the content from the University of Waterloo\'s Desire2Learn LMS',
    author='Stephen Holiday',
    author_email='stephen.holiday@gmail.com',
    url='https://github.com/sholiday/desire2download',
    py_modules=['desire2download'],
    install_requires=(
        'mechanize',
        'BeautifulSoup',
   ),
   scripts=['d2d.py'],
   entry_points={
       'console_scripts': [
           'd2d = d2d:main',
       ]
   },
   classifiers=[
         'Development Status :: 3 - Alpha',
         'Environment :: Console',
         'Intended Audience :: Developers',
         'Intended Audience :: Education',
         'Intended Audience :: End Users/Desktop',
         'License :: OSI Approved :: Apache Software License',
         'Operating System :: MacOS :: MacOS X',
         'Operating System :: POSIX',
         'Operating System :: Unix',
         'Programming Language :: Python :: 2.6',
         'Programming Language :: Python :: 2.7',
         'Programming Language :: Python',
         'Topic :: Education',
   ],
   long_description="""

   ## Note ##
   UWaterloo recently updated Desire2Learn to 10.1 which broke a lot of functionality and introduced bugs.
   This update tries to mitigate some of them but there are a lot of them.
   The main issue is that the folder structure is not preserved and sub folders are not downloaded.

   Download all of the content from the University of Waterloo's
   new learning management system which uses Desire2Learn instead of the old Angel
   based UWACE.

   d2d was inspired by Jamie Wong's fabulous [UWAngel-CLI](https://github.com/phleet/UWAngel-CLI)
   written in Ruby.

   d2d is somewhat hacky and has not been tested extensively. If you do find a bug,
   please [let us know](https://github.com/sholiday/desire2download/issues)

   To install, just do either:

       - python setup.py install
       - pip install desire2download

   To use d2d, just browse to the folder you want to download the files in, type
   d2d and hit enter! d2d will not download a file if it has been already saved.

   Examples:
       - d2d
       - d2d --username ktalwar
       - d2d -u ktalwar -p icanhazcatz
       - d2d -i ".*.ppt"
       - d2d -u scholida -p hecanhazcatz -i ".*.ppt" -c "CS+"

   Result:
       Username: scholida
       Password:
       Logging In...
       Logged In
       Finding courses...
       ECE 224 - Fall 2011
        + ECE 224 - Fall 2011/Labs/Lab Tools Tutorial.html (1.70K)
        + ECE 224 - Fall 2011/Labs/Lab 1/lab1_checklist-s2010.pdf (107.65K)
        ...

   Credits: Stephen Holiday, Ansis Brammanis, Kartik Talwar and Jacob Parry
   """
)



#!/usr/bin/env python
# encoding: utf-8
"""
d2d.py

Copyright 2012 Stephen Holiday

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

# Desire2Download

Download all of the content from the University of Waterloo's
new learning management system which uses Desire2Learn instead of the old Angel
based UWACE.

d2d was inspired by Jamie Wong's fabulous [UWAngel-CLI](https://github.com/phleet/UWAngel-CLI)
written in Ruby.

d2d is somewhat hacky and has not been tested extensively. If you do find a bug,
please [let us know](https://github.com/sholiday/desire2download/issues)

To install, just do either:

   - python setup.py install
   - pip install desire2download

To use d2d, just browse to the folder you want to download the files in, type
d2d and hit enter! d2d will not download a file if it has been already saved.

Examples:
   - d2d
   - d2d --username ktalwar
   - d2d -u ktalwar -p icanhazcatz
   - d2d -i ".*.ppt"
   - d2d -u scholida -p hecanhazcatz -i ".*.ppt" -c "CS+"

Result:
   Username: scholida
   Password:
   Logging In...
   Logged In
   Finding courses...
   ECE 224 - Fall 2011
    + ECE 224 - Fall 2011/Labs/Lab Tools Tutorial.html (1.70K)
    + ECE 224 - Fall 2011/Labs/Lab 1/lab1_checklist-s2010.pdf (107.65K)
      ...

Credits: Stephen Holiday, Ansis Brammanis, Kartik Talwar and Jacob Parry
"""

import getopt
import re

from desire2download import Desire2Download
from desire2download import AuthError
from getpass import getpass
import sys
reload(sys) 
sys.setdefaultencoding("utf-8")

help_message = '''
Desire2Download
===============

Download all of the content from the University of Waterloo's
new learning management system which uses Desire2Learn instead of the old Angel
based UWACE.

d2d was inspired by Jamie Wong's fabulous [UWAngel-CLI](https://github.com/phleet/UWAngel-CLI)
written in Ruby.

d2d is somewhat hacky and has not been tested extensively. If you do find a bug,
please [let us know](https://github.com/sholiday/desire2download/issues)

To use d2d, just browse to the folder you want to download the files in, type
d2d and hit enter! d2d will not download a file if it has been already saved.

Examples:
    - d2d
    - d2d --username ktalwar
    - d2d -u ktalwar -p icanhazcatz
    - d2d -i ".*.ppt"
    - d2d -u scholida -p hecanhazcatz -i ".*.ppt" -c "CS+"

Options:
    -h  --help                  This help message
    -u, --username [username]   set your username
    -p, --password [password]   set your password
    -i, --ignore  [regular exp] ignore files that match this regex
    -c, --courses [regular exp] ignore courses that match this regex
    -o, --overwrite             Overwrite existing files
'''
            
class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg


def main(argv=None):
    if argv is None:
        argv = sys.argv
    try:
        try:
            opts, args = getopt.getopt(argv[1:], "houpi:c:v", ["help", "username=", "password=", "ignore=", "courses=", "overwrite"])
        except getopt.error, msg:
            raise Usage(msg)
            
        username = None
        password = None
        ignore_re = list()
        ignore_course = list()
        skip_existing_files = True
        
        # option processing
        for option, value in opts:
            if option == "-v":
                verbose = True
            if option in ("-h", "--help"):
                raise Usage(help_message)
            if option in ("-u", "--username"):
                username = value
            if option in ("-p", "--password"):
                password = value
            if option in ("-o", "--overwrite"):
                skip_existing_files = False
            if option in ("-i", "--ignore"):
                try:
                    ignore_re.append(re.compile(value))
                except:
                    print 'Regular Expression "%s" is invalid...'
                    raise Usage(help_message)
            if option in ("-c", "--courses"):
                try:
                    ignore_course.append(re.compile(value))
                except:
                    print 'Regular Expression "%s" is invalid...'
                    raise Usage(help_message)
                
        if username is None:
            username = raw_input('Username: ')
        if password is None:
            password = getpass()
        
        
        # Start the actual work
        d2d = Desire2Download(username, password, ignore_re=ignore_re, skip_existing=skip_existing_files)
        try:
            d2d.login()
        except (AuthError, Exception) as e:  ## TODO: replace Exception
            print e
            return 2
        links = d2d.get_course_links()
        trees = []
        for link in links:
            is_skip = False # Use a flag to avoid nasty loop-breaking
            for r in ignore_course:
                if r.match(link.text) is not None:
                    print 'Skipping %s because it matches regex "%s"' % (link.text, r.pattern)
                    is_skip = True
            
            if not is_skip:
                print link.text
                #try:
                trees.append(d2d.get_course_documents(link, link.text))
                #except Exception as e: ## TODO: replace Exception
                #    print 'Failed to load course:', e
                #    #return 2 #don't want to fail just because of one course
        for tree in trees:
            d2d.download_tree(tree)

    except Usage, err:
        print >> sys.stderr, sys.argv[0].split("/")[-1] + ": " + str(err.msg)
        print >> sys.stderr, "\t for help use --help"
        return 2


if __name__ == "__main__":
    sys.exit(main())
